"""Defines the ``Component`` class and ``@component`` decorator."""

from __future__ import annotations

import inspect
from typing import Callable, Type, final

from syntags.lib.syntax import Syntax, SyntaxMeta
from syntags.lib.utils import render, make_repr, flatten


class ComponentMeta(SyntaxMeta):
    """Implements sound __repr__ for non-instantiated Components."""

    def __repr__(cls) -> str:
        return make_repr(cls.__name__, "component")


class Component(Syntax, metaclass=ComponentMeta):
    def __repr__(self) -> str:
        name = self.__class__.__name__
        kids = len(self.children)
        return make_repr(name, "element", children=kids, attrs=self.attrs)

    @final
    def _collapse(self) -> str:
        """Build this component and render the resulting markup."""
        return render(self.build())

    def build(self):
        """Build a tree of nodes."""
        raise NotImplementedError("Component subclasses must implement build()")


def get_component_build_method(builder: Callable) -> Callable:
    """Get a build method based on the signature of the builder callable."""
    params = inspect.signature(builder).parameters

    if next(iter(params.values())).kind == inspect.Parameter.POSITIONAL_ONLY:
        # The first argument is positional-only, so that's now the children.
        def build(self):
            kids = tuple(flatten(self.children))
            return builder(kids, **self.attrs)

    elif "children" in params:
        # The "children" argument is special if there's no positional-only
        # parameters.
        def build(self):
            kids = tuple(flatten(self.children))
            return builder(children=kids, **self.attrs)

    else:
        # None of the above exceptions are met, so call the builder with
        # attributes as keyword arguments.
        def build(self):
            return builder(**self.attrs)

    return build


def component(builder: Callable) -> Type[Component]:
    """Create a Component using the defined builder function."""

    # Use a generic name if builder is a lambda, otherwise use its name
    if builder.__name__ == "<lambda>":
        name = "<lambda component>"
    else:
        name = builder.__name__

    cls_dict = {
        "__module__": builder.__module__,
        "__doc__": builder.__doc__,
        "build": get_component_build_method(builder),
    }

    return type(name, (Component,), cls_dict)



"""Implements the <?xml?> prolog, exports function ``x`` to create XML elements.

Elements can be defined inline or bound to local names. ``x()`` uses
``functools.lru_cache()``, meaning ``x("example")`` will always return the same object.
Binding to local names may be helpful when defining a custom namespace.

Examples of the two methods are below.

    >>> tree = x("root") [
    ...     x("element") [
    ...         x("myThing"),
    ...         x("leafNode")
    ...     ]
    ... ]

    >>> root = x("root")
    >>> element = x("element")
    >>> myThing = x("myThing")
    >>> leafNode = x("leafNode")
    >>> tree = root [
    ...     element [
    ...         myThing,
    ...         leafNode
    ...     ]
    ... ]
"""

__all__ = ["xml", "x"]

from syntags.lib.elements import XMLElement, xml_element as x


class xml(XMLElement):
    _rhs_template = ""
    _lhs_template = _void_template = "<?$name?>"
    _lhs_attr_template = _void_attr_template = "<?$name $attrs?>"



"""XML tags for the sitemap schema."""

__all__ = [
    "changefreq",
    "lastmod",
    "loc",
    "priority",
    "sitemap",
    "sitemapindex",
    "url",
    "urlset",
]

from syntags.lib.elements import xml_element
from syntags.lib.components import component
from syntags.ext import xml_prolog as xml

SITEMAP_NS = "http://www.sitemaps.org/schemas/sitemap/0.9"


@component
def urlset(children, **attrs):
    """Create a <urlset> element with preceeding <?xml?> prolog."""
    return xml / xml_element("urlset")(attrs, xmlns=SITEMAP_NS)[children]


@component
def sitemapindex(children, **attrs):
    """Create a <sitemapindex> element with preceeding <?xml?> prolog."""
    return xml / xml_element("sitemapindex")(attrs, xmlns=SITEMAP_NS)[children]


changefreq = xml_element("changefreq")
lastmod = xml_element("lastmod")
loc = xml_element("loc")
priority = xml_element("priority")
sitemap = xml_element("sitemap")
url = xml_element("url")



"""Implements utilities relied on by other modules in the package.

This file should not import anything locally, because that will probably lead to a
circular import.
"""

from __future__ import annotations

import collections.abc
from functools import singledispatch
from textwrap import dedent
from os import PathLike
from typing import Any, Optional, Sequence
from xml.sax.saxutils import escape, quoteattr as quote

ATTR_ALIASES = {"classes": "class", "for_id": "for", "is_async": "async"}


class RawSentinel:
    """Empty sentinel mixin to indicate a type should be treated as safe when str'd."""


class RawStr(str, RawSentinel):
    """String that will not be escaped when rendered."""


def raw(text) -> RawStr:
    """Dedent the text and make it raw."""
    string = str(text)
    dedented = dedent(string)
    stripped = dedented.strip()
    return RawStr(stripped)


def flatten(fragment: Sequence[Any]):
    """Flatten a fragment of any depth."""
    for node in fragment:
        if isinstance(node, Sequence) and not isinstance(node, str):
            # node is actually a fragment itself, flatten it!
            yield from flatten(node)
        else:
            yield node


def render(markup) -> str:
    """Convert any given markup to a string representation."""
    # Make sure iterating over markup will be correct
    if isinstance(markup, str) or not isinstance(markup, Sequence):
        markup = (markup,)

    sink = []
    for x in flatten(markup):
        # None is used to represent no content, and ... is for documentation.
        if x in (None, ...):
            continue

        # Anything subclassing RawSentinel should not be escaped.
        if isinstance(x, RawSentinel):
            sink.append(str(x))

        # Anything else should be escaped.
        else:
            sink.append(escape(str(x)))

    return "".join(sink)


def render_to_file(path: PathLike, markup) -> None:
    """Write rendered markup to a file."""

    with open(path, "w") as f:
        f.write(render(markup))


def render_attrs(attr_dict) -> str:
    """Render the dictionary to a valid HTML space-separated attribute list."""
    rendered = (render_attr(val, name) for name, val in attr_dict.items())
    return " ".join(s for s in rendered if s)


@singledispatch
def render_attr(val, name: str) -> Optional[str]:
    """Render an attribute based on its value.

    Dispatches to the appropriate function.
    """
    return f"{name}={quote(str(val))}"


# RawSentinel should always be literal, must be first.
@render_attr.register(RawSentinel)
def _(val, name):
    return f'{name}="{val}"'


@render_attr.register(str)
def _(val, name):
    return f"{name}={quote(val)}"


@render_attr.register(bool)
def _(val, name):
    return name if val else None


# see: https://bugs.python.org/issue34498
@render_attr.register(collections.abc.Sequence)
def _(val, name):
    joined = " ".join(str(x) for x in val)
    return f"{name}={quote(joined)}"


# see: https://bugs.python.org/issue34498
@render_attr.register(collections.abc.Iterable)
def _(val, name):
    as_strings = sorted(map(str, val))
    joined = " ".join(as_strings)
    escaped = quote(joined)
    return f"{name}={escaped}"


def make_repr(*info, **details) -> str:
    """Concatenate information into a readable repr."""
    items = (str(arg) for arg in (*info, details) if arg)
    return "<" + " ".join(items) + ">"



"""Namespace of SVG tags.

Tag names match their canonical web equivalents in casing unless they shadow a builtin,
in which case the first letter will be upper-cased.
"""

__all__ = [
    "a",
    "animate",
    "animateMotion",
    "animateTransform",
    "audio",
    "canvas",
    "circle",
    "clipPath",
    "defs",
    "desc",
    "discard",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "Filter",
    "foreignObject",
    "g",
    "iframe",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "script",
    "Set",
    "stop",
    "style",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "title",
    "tspan",
    "unknown",
    "use",
    "video",
    "view",
]

from syntags.lib.elements import xml_element

# From:
#   https://www.w3.org/TR/SVG/eltindex.html

a = xml_element("a")
animate = xml_element("animate")
animateMotion = xml_element("animateMotion")
animateTransform = xml_element("animateTransform")
audio = xml_element("audio")
canvas = xml_element("canvas")
circle = xml_element("circle")
clipPath = xml_element("clipPath")
defs = xml_element("defs")
desc = xml_element("desc")
discard = xml_element("discard")
ellipse = xml_element("ellipse")
feBlend = xml_element("feBlend")
feColorMatrix = xml_element("feColorMatrix")
feComponentTransfer = xml_element("feComponentTransfer")
feComposite = xml_element("feComposite")
feConvolveMatrix = xml_element("feConvolveMatrix")
feDiffuseLighting = xml_element("feDiffuseLighting")
feDisplacementMap = xml_element("feDisplacementMap")
feDistantLight = xml_element("feDistantLight")
feDropShadow = xml_element("feDropShadow")
feFlood = xml_element("feFlood")
feFuncA = xml_element("feFuncA")
feFuncB = xml_element("feFuncB")
feFuncG = xml_element("feFuncG")
feFuncR = xml_element("feFuncR")
feGaussianBlur = xml_element("feGaussianBlur")
feImage = xml_element("feImage")
feMerge = xml_element("feMerge")
feMergeNode = xml_element("feMergeNode")
feMorphology = xml_element("feMorphology")
feOffset = xml_element("feOffset")
fePointLight = xml_element("fePointLight")
feSpecularLighting = xml_element("feSpecularLighting")
feSpotLight = xml_element("feSpotLight")
feTile = xml_element("feTile")
feTurbulence = xml_element("feTurbulence")
Filter = xml_element("filter")
foreignObject = xml_element("foreignObject")
g = xml_element("g")
iframe = xml_element("iframe")
image = xml_element("image")
line = xml_element("line")
linearGradient = xml_element("linearGradient")
marker = xml_element("marker")
mask = xml_element("mask")
metadata = xml_element("metadata")
mpath = xml_element("mpath")
path = xml_element("path")
pattern = xml_element("pattern")
polygon = xml_element("polygon")
polyline = xml_element("polyline")
radialGradient = xml_element("radialGradient")
rect = xml_element("rect")
script = xml_element("script")
Set = xml_element("set")
stop = xml_element("stop")
style = xml_element("style")
svg = xml_element("svg")
switch = xml_element("switch")
symbol = xml_element("symbol")
text = xml_element("text")
textPath = xml_element("textPath")
title = xml_element("title")
tspan = xml_element("tspan")
unknown = xml_element("unknown")
use = xml_element("use")
video = xml_element("video")
view = xml_element("view")



"""Namespace for HTML tags.

Tag names match their canonical web equivalents in casing unless they shadow a builtin,
in which case the first letter will be upper-cased.
"""

__all__ = [
    "a",
    "abbr",
    "acronym",
    "address",
    "applet",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "basefont",
    "bdi",
    "bdo",
    "big",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "center",
    "cite",
    "code",
    "col",
    "colgroup",
    "command",
    "comment",
    "data",
    "datalist",
    "dd",
    "Del",
    "details",
    "dfn",
    "dialog",
    "Dir",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "font",
    "footer",
    "form",
    "frame",
    "frameset",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hgroup",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "Input",
    "ins",
    "isindex",
    "kbd",
    "keygen",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "Map",
    "mark",
    "menu",
    "menuitem",
    "meta",
    "meter",
    "nav",
    "nextid",
    "noframes",
    "noscript",
    "Object",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "param",
    "picture",
    "plaintext",
    "pre",
    "progress",
    "q",
    "rb",
    "rbc",
    "rp",
    "rt",
    "rtc",
    "ruby",
    "s",
    "samp",
    "script",
    "section",
    "select",
    "small",
    "source",
    "span",
    "strike",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "table",
    "tbody",
    "td",
    "template",
    "textarea",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "track",
    "tt",
    "u",
    "ul",
    "var",
    "video",
    "wbr",
    "xmp",
]

from syntags.lib.elements import element, Element

# List from:
#   http://w3c.github.io/elements-of-html/
#   https://www.w3.org/TR/2012/WD-html-markup-20121025/elements.html
# Void elements as seen in:
#   https://www.w3.org/TR/2011/WD-html-markup-20110405/syntax.html#syntax-elements
#   https://developer.mozilla.org/en-US/docs/Glossary/empty_element

a = element("a")
abbr = element("abbr")
acronym = element("acronym")
address = element("address")
applet = element("applet")
area = element("area", void=True)
article = element("article")
aside = element("aside")
audio = element("audio")
b = element("b")
base = element("base", void=True)
basefont = element("basefont", void=True)
bdi = element("bdi")
bdo = element("bdo")
big = element("big")
blockquote = element("blockquote")
body = element("body")
br = element("br", void=True)
button = element("button")
canvas = element("canvas")
caption = element("caption")
center = element("center")
cite = element("cite")
code = element("code")
col = element("col", void=True)
colgroup = element("colgroup")
command = element("command", void=True)
data = element("data")
datalist = element("datalist")
dd = element("dd")
Del = element("del", void=True)
details = element("details")
dfn = element("dfn")
dialog = element("dialog")
Dir = element("dir", void=True)
div = element("div")
dl = element("dl")
dt = element("dt")
em = element("em")
embed = element("embed", void=True)
fieldset = element("fieldset")
figcaption = element("figcaption")
figure = element("figure")
font = element("font")
footer = element("footer")
form = element("form")
frame = element("frame", void=True)
frameset = element("frameset")
h1 = element("h1")
h2 = element("h2")
h3 = element("h3")
h4 = element("h4")
h5 = element("h5")
h6 = element("h6")
head = element("head")
header = element("header")
hgroup = element("hgroup")
hr = element("hr", void=True)
i = element("i")
iframe = element("iframe")
img = element("img", void=True)
Input = element("input", void=True)
ins = element("ins")
isindex = element("isindex", void=True)
kbd = element("kbd")
keygen = element("keygen", void=True)
label = element("label")
legend = element("legend")
li = element("li")
link = element("link", void=True)
main = element("main")
Map = element("map", void=True)
mark = element("mark")
menu = element("menu")
menuitem = element("menuitem", void=True)
meta = element("meta", void=True)
meter = element("meter")
nav = element("nav")
nextid = element("nextid", void=True)
noframes = element("noframes")
noscript = element("noscript")
Object = element("object", void=True)
ol = element("ol")
optgroup = element("optgroup")
option = element("option")
output = element("output")
p = element("p")
param = element("param", void=True)
picture = element("picture")
plaintext = element("plaintext")
pre = element("pre")
progress = element("progress")
q = element("q")
rb = element("rb")
rbc = element("rbc")
rp = element("rp")
rt = element("rt")
rtc = element("rtc")
ruby = element("ruby")
s = element("s")
samp = element("samp")
script = element("script")
section = element("section")
select = element("select")
small = element("small")
source = element("source", void=True)
span = element("span")
strike = element("strike")
strong = element("strong")
style = element("style")
sub = element("sub")
summary = element("summary")
sup = element("sup")
table = element("table")
tbody = element("tbody")
td = element("td")
template = element("template")
textarea = element("textarea")
tfoot = element("tfoot")
th = element("th")
thead = element("thead")
time = element("time")
title = element("title")
tr = element("tr")
track = element("track", void=True)
tt = element("tt")
u = element("u")
ul = element("ul")
var = element("var")
video = element("video")
wbr = element("wbr", void=True)
xmp = element("xmp")


class html(Element):
    _name = "html"
    _lhs_template = "<!DOCTYPE $name><$name>"
    _lhs_attr_template = "<!DOCTYPE $name><$name $attrs>"


class comment(Element):
    _lhs_template = "<!-- "
    _lhs_attr_template = "<!-- "
    _rhs_template = " -->"



"""Defines most of the public interface for Syntags.

Attributes on ``Element`` and ``Component`` all start with underscores. This keeps
more names free to be used by the HTML class syntax.
"""

from __future__ import annotations

from functools import lru_cache
from string import Template
from typing import ClassVar, Optional, Tuple, TypeVar

from syntags.lib.syntax import Syntax, SyntaxMeta
from syntags.lib.utils import make_repr, render, render_attrs

T = TypeVar("T")


class ElementMeta(SyntaxMeta):
    """Implements __repr__ for non-instantiated Elements."""

    def __repr__(cls) -> str:
        return make_repr(cls()._name, "element")


class Element(Syntax, metaclass=ElementMeta):
    """A Syntax type that can render itself."""

    _name: str

    # If there are no children, should the tag render as a void tag?
    _void_if_leaf: ClassVar[bool] = False

    _lhs_template: ClassVar[str] = "<$name>"
    _lhs_attr_template: ClassVar[str] = "<$name $attrs>"
    _void_template: ClassVar[str] = "<$name>"
    _void_attr_template: ClassVar[str] = "<$name $attrs>"
    _rhs_template: ClassVar[str] = "</$name>"

    @classmethod
    @lru_cache(None)
    def _new(cls: T, name: str, *, void: Optional[bool] = None) -> T:
        """Create a subclass of this class using the given name."""

        is_void = {} if void is None else {"_void_if_leaf": void}
        cls_dict = {"_name": name, **is_void}
        return type(name, (cls,), cls_dict)

    @property
    def _name(self) -> str:
        """Get the name used in template substitution."""
        return self.__class__.__name__

    def _render_lhs(self) -> str:
        """Render the left side of the tag, with attributes if present."""
        if not self.attrs:
            lhs = Template(self._lhs_template)
            return lhs.safe_substitute(name=self._name)

        lhs = Template(self._lhs_attr_template)
        attrs = render_attrs(self.attrs)
        return lhs.safe_substitute(name=self._name, attrs=attrs)

    def _render_rhs(self) -> str:
        """Render the closing tag."""
        rhs = Template(self._rhs_template)
        return rhs.safe_substitute(name=self._name)

    def _render_as_void(self) -> str:
        """Render the complete tag, with attributes if present.

        This method is only called if both ``self._void_if_leaf == True`` and
        there are no children, otherwise the tag is rendered normally.
        """
        if not self.attrs:
            void = Template(self._void_template)
            return void.safe_substitute(name=self._name)

        void = Template(self._void_attr_template)
        attrs = render_attrs(self.attrs)
        return void.safe_substitute(name=self._name, attrs=attrs)

    def _render_self(self) -> Tuple[str, str]:
        """Render the entire tag, split into open and close strings.

        If the tag is void and has no children, the void tag and an empty string will
        be returned instead of the open and close tags.
        """
        if self._void_if_leaf and not self.children:
            return self._render_as_void(), ""

        return self._render_lhs(), self._render_rhs()

    def _render_children(self) -> str:
        """Render each child of this element."""
        return render(self.children)

    def _collapse(self) -> str:
        """Collapse the element and all its children to a string."""
        lhs, rhs = self._render_self()

        kids = self._render_children()

        return f"{lhs}{kids}{rhs}"

    def __repr__(self) -> str:
        kids = len(self.children)
        return make_repr(self._name, "element", children=kids, attrs=self.attrs)


class XMLElement(Element):
    """Subclass of Element with correct XML semantics."""

    _void_if_leaf = True

    _void_template = "<$name />"
    _void_attr_template = "<$name $attrs />"


element = Element._new

xml_element = XMLElement._new



"""Defines the ``Syntax`` base class, implemented by Element and Component.

``Syntax`` purely implements the syntax and data structure. It has absolutely no
idea what it's used for.
"""

from __future__ import annotations

from functools import reduce
from typing import Any, Dict, Mapping, Optional, Sequence, TypeVar, Union, final

from syntags.lib.utils import ATTR_ALIASES, RawSentinel

T = TypeVar("T", bound="Syntax")


class SyntaxMeta(type, RawSentinel):
    """Component metaclass for implementing shorthand syntax on the type."""

    def __str__(cls) -> str:
        return cls().__str__()

    def __getattr__(cls, name: str) -> Syntax:
        return cls().__getattr__(name)

    def __truediv__(cls, other: Union[Syntax, SyntaxMeta]) -> Syntax:
        return cls().__truediv__(other)

    def __getitem__(cls, others: Any) -> Syntax:
        return cls().__getitem__(others)


class Syntax(RawSentinel, metaclass=SyntaxMeta):
    attrs: Dict[str, Any]
    children: Sequence[Any]

    _proxy: Syntax

    # ——— Setting attributes ———

    def __init__(self, *attr_dicts: Optional[Mapping[str, Any]], **attrs: Any):
        """Optionally set attrs using either a dict (literal) or keywords (aliased)."""
        self._proxy = self
        self.children = ()
        self.attrs = {}

        # all the logic is already in __call__, no need to repeat :)
        self(*attr_dicts, **attrs)

    def __call__(self: T, *attr_dicts: Optional[Mapping[str, Any]], **attrs: Any) -> T:
        """Merge new attributes with any previously set, using a dict or keywords."""

        # Must be first so it can be overridden by keyword args.
        attr_union = reduce(lambda cum, cur: {**cum, **cur}, attr_dicts, {})
        self.attrs.update(attr_union)

        # Use aliases and replace underscores with dashes so users rarely
        # need to use attr_dict
        for key, val in attrs.items():
            safe_key = ATTR_ALIASES.get(key, key)
            self.attrs[safe_key.replace("_", "-")] = val

        return self

    # ——— Setting a proxy child ———

    def _with_child(self: T, child: Union[Syntax, SyntaxMeta]) -> T:
        """Set a child node to use a proxy when adding future children."""
        if isinstance(child, SyntaxMeta):
            child = child()

        if not isinstance(child, Syntax):
            msg = "Combination is only valid between two Syntax subclasses, not with "
            msg += repr(type(child).__name__)
            raise TypeError(msg)

        self._around(child)
        self._proxy = child
        return self

    def __truediv__(self: T, child: Union[Syntax, SyntaxMeta]) -> T:
        return self._with_child(child)

    # ——— Adding children ———

    def _around(self: T, others: Any) -> T:
        """Add multiple children to this node."""
        # The proxy is normally ``self``, but is sometimes a child. This allows
        # combining nodes by giving new children to the proxy.
        if isinstance(others, str) or not isinstance(others, Sequence):
            self._proxy.children = (others,)
        else:
            self._proxy.children = others
        return self

    def __getitem__(self: T, others: Any) -> T:
        return self._around(others)

    # ——— Adding classes ———

    def _add_class(self: T, name: str) -> T:
        """Merge a new name with previously set names, or create a set with one name.

        Note that underscores will be replaced by dashes in the name. This conflicts
        with BEM naming -- use the ``classes`` keyword or ``class`` dict key.
        """
        if name.startswith("_"):
            return self

        fixed_format = name.replace("_", "-")

        classes = self.attrs.setdefault("class", set())

        if isinstance(classes, str):
            # Splits on spaces to create a set, ensures unique classes
            self.attrs["class"] = {*classes.split(), fixed_format}
        else:
            classes.add(fixed_format)

        return self

    def __getattr__(self: T, name: str) -> T:
        return self._add_class(name)

    # ——— Rendering logic ———

    def _collapse(self) -> str:
        """Create a human-friendly string representation."""
        raise NotImplementedError("Syntax subclasses must implement _collapse()")

    @final
    def __str__(self) -> str:
        return self._collapse()

    # ——— Equality & Comparison ———

    def __eq__(self, other) -> bool:
        return self.__class__ is other.__class__ and self.attrs == other.attrs



"""Various components that provide non-standard functionality.

These components simplify common repetitive/verbose elements, such as adding CSS,
fonts from Google Fonts, or elements that can only ever have -- and must contain -- one
child (looking at <rss> 👀).

Components import their own dependencies. This avoids evaluating each namespace
when this file is imported.
"""

__all__ = ["css", "google_fonts", "xml_prolog", "rss_channel", "markdown"]

from syntags.lib.components import component
from syntags.lib.utils import RawStr

# fmt: off


@component
def css(**attrs):
    """Build a link containing the expected rel and type attributes."""
    from syntags.html import link

    return link (attrs, rel="stylesheet", type="text/css")


@component
def google_fonts(fonts, **attrs):
    """Build a link with the correct Google Fonts URL.

    Must use the ``fonts`` attribute with an iterable (list, set, etc).

        >>> google_fonts (fonts={"Roboto", "Roboto Slab:600,700"})

    Additional attributes will be passed to the created ``link`` element.
    """
    base_url = "https://fonts.googleapis.com/css?display=swap&family="
    formatted = "|".join(fonts).replace(" ", "+")
    fonts_url = base_url + formatted
    return css (attrs, href=fonts_url)


@component
def xml_prolog(children, **attrs):
    """Create an XML prolog with UTF-8 encoding and the version set to 1.0.

        >>> from syntags.xml import x
        >>> from syntags.ext import xml_prolog
        >>> xml_prolog / x("root") [
        ...     x("my-element") [
        ...         "example"
        ...     ]
        ... ]
    """
    from syntags.xml import xml

    return xml (attrs, version="1.0", encoding="UTF-8") [children]


@component
def rss_channel(children, **attrs):
    """Create a channel, with RSS 2.0 tag and XML 1.0 prolog.

        >>> from syntags.rss import *
        >>> from syntags.ext import rss_channel
        >>> rss_channel [
        ...     title ["A good example"],
        ...     author ["SeparateRecords"],
        ...     ...
        ... ]
    """
    from syntags.rss import rss, channel

    return xml_prolog / rss (attrs, version="2.0") / channel [children]


@component
def markdown(children, extras=None, src=None, **attrs):
    """Render some markdown to a raw HTML string.

    This component uses markdown2 for rendering. It enables the
    fenced-code-blocks and header-ids extras, and allows you to add more with
    the ``extras`` attribute (list).

    See the following link for a list of the supported extras:
    https://github.com/trentm/python-markdown2/wiki/Extras#implemented-extras

    You can render a file.

        >>> markdown (src="posts/1.md")

    Or some inline markdown.

        >>> markdown [
        ...     '''
        ...     # Title
        ...
        ...     Content
        ...     '''
        ... ]

    Or both, using the ellipsis literal as a placeholder.

        >>> markdown (src="docs/1-syntax.md") [
        ...     ...,
        ...     '''
        ...
        ...     ## Style Recommendations
        ...
        ...     See the [styleguide] for recommendations on how to keep your
        ...     code readable and maintainable in the long term, and configuration
        ...     for your linter.
        ...
        ...     '''
        ... ]

    This component requires the "markdown" extra.

        $ pip install syntags[markdown]

    """
    import textwrap

    try:
        import markdown2 as md
    except ImportError as e:
        raise ImportError("Syntags: Missing feature 'markdown'") from e

    # Add some sensible extras to control how the markdown renders.
    md_extras = ["fenced-code-blocks", "header-ids"]
    if extras:
        md_extras += extras

    # Only render path to markdown if it's actually needed.
    rendered_src = ""

    if src and (not children or ... in children):
        rendered_src = md.markdown_path(src, extras=md_extras)

    if not children:
        return RawStr(rendered_src)

    # Each item is a sequence of uninterrupted, non-ellipsis children that
    # have been joined by newlines.
    joined_md = []

    # Collects children until an ellipsis is found and they get moved to
    # ``joined_md``
    sink = []
    for child in children:
        if child is not ...:
            as_string = str(child)
            dedented = textwrap.dedent(as_string)
            stripped = dedented.strip()
            sink.append(stripped)
        else:  # child is ...
            joined_md.append("\n".join(sink))
            sink.clear()

    # There's still stuff in the sink if there was no ellipsis
    joined_md.append("\n".join(sink))

    rendered = [md.markdown(item, extras=md_extras) for item in joined_md]
    final_markdown = rendered_src.join(rendered)

    return RawStr(final_markdown)



"""RSS 2.0 tags.

See ``tagsext`` for useful components to simplify some boilerplate.

Tag names match their canonical web equivalents in casing unless they shadow a builtin,
in which case the first letter will be upper-cased.
"""

__all__ = [
    "author",
    "category",
    "channel",
    "cloud",
    "comments",
    "Copyright",
    "day",
    "description",
    "docs",
    "enclosure",
    "generator",
    "guid",
    "height",
    "hour",
    "image",
    "item",
    "language",
    "lastBuildDate",
    "link",
    "managingEditor",
    "name",
    "pubDate",
    "rating",
    "rss",
    "skipDays",
    "skipHours",
    "source",
    "textInput",
    "title",
    "ttl",
    "url",
    "webMaster",
    "width",
    "xml",
]

from syntags.lib.elements import xml_element
from syntags.xml import xml

author = xml_element("author")
category = xml_element("category")
channel = xml_element("channel")
cloud = xml_element("cloud")
comments = xml_element("comments")
Copyright = xml_element("copyright")
day = xml_element("day")
description = xml_element("description")
docs = xml_element("docs")
enclosure = xml_element("enclosure")
generator = xml_element("generator")
guid = xml_element("guid")
height = xml_element("height")
hour = xml_element("hour")
image = xml_element("image")
item = xml_element("item")
language = xml_element("language")
lastBuildDate = xml_element("lastBuildDate")
link = xml_element("link")
managingEditor = xml_element("managingEditor")
name = xml_element("name")
pubDate = xml_element("pubDate")
rating = xml_element("rating")
rss = xml_element("rss")
skipDays = xml_element("skipDays")
skipHours = xml_element("skipHours")
source = xml_element("source")
textInput = xml_element("textInput")
title = xml_element("title")
ttl = xml_element("ttl")
url = xml_element("url")
webMaster = xml_element("webMaster")
width = xml_element("width")



"""Write your markup in pure Python.

See the readme on GitHub for documentation, examples, and other information.

https://github.com/SeparateRecords/Syntags
"""

#
# Copyright (c) 2020, Robert "SeparateRecords" Cooper <me@rob.ac>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

__title__ = "syntags"
__version__ = "0.1.0"
__author__ = "Robert Cooper <me@rob.ac>"
__copyright__ = "Copyright 2020, Robert Cooper (SeparateRecords)"
__license__ = "ISC"
__all__ = (
    # ~~ Public API ~~ The intended interface for end users.
    "component",
    "element",
    "raw",
    "render",
    "render_to_file",
    # ~~ Tag Namespaces ~~ These aren't imported to avoid evaluating them.
    "html",
    "xml",
    "rss",
    "sitemap",
    "svg",
    "ext",
    # ~~ Raw API ~~ Exposes some of the inner workings of the API. For nerds.
    "RawStr",
    "Component",
    "Element",
    "ATTR_ALIASES",
    # ~~ API ABCs ~~ For even bigger nerds.
    "RawSentinel",
    "Syntax",
    "SyntaxMeta",
)

from syntags.lib.components import Component, component
from syntags.lib.elements import Element, XMLElement, element, xml_element
from syntags.lib.syntax import Syntax, SyntaxMeta
from syntags.lib.utils import (
    ATTR_ALIASES,
    RawStr,
    RawSentinel,
    raw,
    render,
    render_to_file,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['syntags', 'syntags.lib']

package_data = \
{'': ['*']}

extras_require = \
{'markdown': ['markdown2>=2.3.8,<3.0.0', 'pygments>=2.5.2,<3.0.0']}

setup_kwargs = {
    'name': 'syntags',
    'version': '0.1.0',
    'description': 'A beautiful python syntax to generate HTML.',
    'long_description': '<div align="center">\n\n# Syntags: Write your markup in Python, easily\n\n[Features] &emsp;•&emsp; [Install] &emsp;•&emsp; [License] &emsp;•&emsp; [Docs]\n\n</div>\n\n[Features]: #features\n[Install]: #install\n[License]: #license\n[Docs]: https://github.com/SeparateRecords/Syntags/tree/master/docs/\n\n**Syntags** lets you generate HTML, XML, SVG, etc. by writing concise, _real_ Python code.\n\n**Syntags** was built from the ground up for readable code. First-class support for _components_, _custom elements_, and _fragments_ is baked into the design.\n\n**Syntags** takes a different approach to Pyxl — taking Python\'s own syntax and repurposing it, rather than using a custom character encoding and transforming your invalid code into valid code.\n\n* The code you write is **real** Python, no magic.\n* You don\'t need to learn the quirks of a new syntax.\n* It\'s easy to get started with and integrate into your code.\n* Highlighting works, and your linter will only be _slightly_ angry.\n\n#### A brief note on the syntax ~\n\nPEP 8 isn\'t gospel. Creating a beautiful syntax means ignoring certain parts of it, and that\'s okay! The second section, "[A Foolish Consistency is the Hobgoblin of Little Minds][consistency]," even summarises this.\n\n<details>\n<summary>View a relevant excerpt</summary>\n\n> . . . **know when to be inconsistent** -- sometimes style guide\n> recommendations just aren\'t applicable. When in doubt, use your best judgment.\n> Look at other examples and **decide what looks best**.\n>\n> Some other good reasons to ignore a particular guideline:\n>\n> 1. When **applying the guideline would make the code less readable**, even for\n>    someone who is used to reading code that follows this PEP.\n>\n> 2. . . .\n\n</details>\n\nCheck out the [Syntags styleguide][styleguide] for recommendations on how to maintain good code quality.\n\n[consistency]: https://www.python.org/dev/peps/pep-0008/#a-foolish-consistency-is-the-hobgoblin-of-little-minds\n\n[styleguide]: https://github.com/SeparateRecords/Syntags/tree/master/docs/styleguide.md\n\n## Features\n\n### an expressive syntax that _makes sense_\n\nStructured like a programming language, because it _is_ one. No context switching between templating and programming.\n\n```python\nimport syntags as tags\nfrom syntags.html import *\n\napp = html (lang="en") [\n  head [\n    title ["Some basic markup"]\n  ],\n  body [\n    main [\n      ...  # etc...\n    ]\n  ]\n]\n\nrendered = tags.render(app)\n```\n\n### intuitive syntax shorthands\n\nIt\'s effortless to write compact code that\'s easy to read and maintain.\n\n```python\nbody [\n  div .main_wrapper / main [\n    ...\n  ]\n]\n\n# The same, but without the shorthand:\nbody [\n  div (classes={"main-wrapper"}) [\n    main [\n      ...\n    ]\n  ]\n]\n```\n\n### reusable, declarative components\n\nSimilar to React, but very much Python.\n\n```python\nimport syntags as tags\nfrom syntags.html import *\n\n@tags.component\ndef hello_message(name, **attrs):\n  return div [\n    f"Hello {name}"\n  ]\n\n\nrv = tags.render(hello_message (name="world"))\n```\n\n<details>\n<summary>View the React JSX equivalent</summary>\n\n```jsx\nimport React from "react";\nimport ReactDOMServer from "react-dom/server";\n\nfunction HelloMessage({ name, ...props }) {\n  return (\n    <div>\n      Hello {name}\n    </div>\n  );\n}\n\n\nrv = ReactDOMServer.renderToStaticMarkup(<HelloMessage name="world" />)\n```\n\nKeep in mind, this is in JSX. Compiling it is yet another step. Syntags\' syntax is pure Python, no intermediate steps required.\n\n</details>\n\n### first-class fragments, no wizardry required\n\nFragments are just a sequence of elements, and they can be nested infinitely.\n\n```python\nHEAD_DATA = [\n  link (rel="canonical", href="https://rob.ac/"),\n  link (rel="stylesheet", href="dist/css/layout.css"),\n  link (rel="shortcut icon", href="favicon.ico", type="image/x-icon")\n]\n\nhtml (lang="en") [\n  head [\n    HEAD_DATA\n  ],\n  ...\n]\n```\n\n### simple, safe rendering\n\nEverything becomes a string and gets escaped, except for raw strings.\n\n```python\nimport syntags as tags\nfrom syntags.html import *\n\nrendered = tags.render(\n  html (lang="en") [\n    tags.raw("""\n    <head>\n      <title>A good example</title>\n      <meta charset="UTF-8">\n    </head>\n    """),\n    body [\n      p [\n        "<b>This will get escaped</b>"\n      ]\n    ]\n  ]\n)\n```\n\n### more than just HTML\n\nSyntags comes with a bunch of namespaces to make your life easy.\n\n* HTML tags: `html`\n* XML prolog and tag factory: `xml`\n* Some predefined XML namespaces: `svg`, `rss`, `sitemap`\n* Common, useful components: `ext`\n\n## Install\n\nYou can use Pip, or anything that can install from PyPI, such as [Poetry] or [Pipenv].\n\n```console\n$ pip install syntags\n```\n\n[Poetry]: https://python-poetry.org\n[Pipenv]: https://pipenv.pypa.io/en/latest/\n\n###### + Markdown component (`syntags.ext.markdown`)\n\n```console\n$ pip install syntags[markdown]\n```\n\n## License\n\nSyntags is licensed under the ISC license, a simplified version of the MIT license. You can modify this project and use it in whatever you want! The only condition for redistribution is that you must include a copy of the license.\n\nThere\'s a copy in `syntags/__init__.py`, so you\'ll be fine as long as you don\'t remove that! Otherwise, you can include the copy found in the root of [the repository][repo].\n\n[repo]: https://github.com/SeparateRecords/Syntags\n',
    'author': 'SeparateRecords',
    'author_email': 'me@rob.ac',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SeparateRecords/Syntags',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
